/* Coral library for low level data structure implementation
 Copyright (C) 1996-2004 Hee S. Chung

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* Text I/O operations */
/*
 1997.10
 1999.06.05
 1999.11.03  Revision using I/O buffer API
 2000.07.26  Reviewed for PARAM spec to support classic param style
 2001.03.14  The definition of I/O buffer
*/

#undef __DEBUG__ /* ignore globally defined debug request */
/* #define __DEBUG__ */



#include <coralsrc/io.h>
#include <coralsrc/memory.h>
#include <coralsrc/string.h>
#include <coral/inline.h>
#include <coral/compat/ctype.h>
#include <coral/compat/string.h>




/*------------------------------------------------------------------------*
   Internal functions
 *------------------------------------------------------------------------*/



__MODULEBEGIN__( "iotdat.c" )
struct _IOTEXT  io_text_default;
__MODULEEND__



/* iotext_skip_lineend

	This function is ( should be ) called after p is increased at least by 1
	*p must be '\r' or '\n' and at least, this function will return p+1.

	This function can modify the given size, so the size after calling
	this function should be carefully evaluated. */

_integer1 __F*
iotext_skip_lineend
	P_B(( iob, p, size ))
	P_1		struct _IOBUFFER __F* iob
	P_N		_integer1 __F* p
	P_N		int __F* size
	P_E
{
	register char c1, c2;
	int i;

	/* evidently c1 should be line end character ( it's a rule for this
	  function )
	   this function is ( should be ) called after p is increased at least
	  by 1 */
	c1 = *(p-1);
	i = *size;

	if( i<=0 ) {
		iob->unlock( iob->context, p );
		p = (_integer1 __F*)iob->lock( iob->context, &i, 0, IO_FLAG_READ );
		if( !p ) return NULL;
		if( i==0 ) return p; /* go on further buffered io operations */
	}

	c2=*p;

	if( ( c2!='\r' && c2!='\n' ) || c1==c2 ) {
		/* the current character is not a line end character or
			 \n\n or \r\r means two lines */
		*size = i;
		return p;
	}

	/* the current character is a line end character and the same
	  line end character was not appeared ( \r\n or \n\r is allowed ) */

	*size = i-1;
	return p+1;
}



/*------------------------------------------------------------------------*
   API
 *------------------------------------------------------------------------*/



/* iotext_compare_string

  The first caracter must be tested before calling this function */

__MODULEBEGIN__( "iotcmps.c" )

int
iotext_compare_string
	P_B(( iob, iot, locked, remained, s, len ))
	P_1		struct _IOBUFFER __F* iob
	P_N		struct _IOTEXT __F* iot
	P_N		_integer1 __F*__F* locked
	P_N		int __F* remained
	P_N		const char __F* s
	P_N		int len
	P_E
{
	_integer1 __F* p=*locked;
	if( len>1 ) {
		int size=*remained;
		if( size<len ) { /* remained locked chars are shorter than
							the current comment */
			iob->unlock( iob->context, p );
			p = iob->lock( iob->context, &size, len, IO_FLAG_READ );
			/*  at least one byte is remained because the first character
			   should be tested before calling this function, so though p is
			   NULL now, the following lock operation will be evidently
			   successful : returning error code will cause more difficult
			   exception handling in caller side */
			if( !p ) p = iob->lock( iob->context, &size, 0, IO_FLAG_READ );
			*locked = p;
			*remained = size;

			/* insufficient locked chars so it is not a comment */
			if( size<len ) return FAIL_UNAVAILABLE;
		}

		if( strncmp((const char __F*)p,s,len)!=0 ) return 1;
	}
	return 0;
}
__MODULEEND__



__MODULEBEGIN__( "iotskip.c" )
/* iotext_skip_comment */

int
iotext_skip_comment
	P_B(( iob, iot, cmt, lcmt, eventflag, charflag,
			 startevent, endevent, charevent, line ))
	P_1		struct _IOBUFFER __F* iob
	P_N		struct _IOTEXT __F* iot
	P_N		const char __F* cmt
	P_N		int lcmt
	P_N		unsigned eventflag
	P_N		unsigned charflag
	P_N		unsigned startevent
	P_N		unsigned endevent
	P_N		unsigned charevent
	P_N		unsigned long __F* line
	P_E
{
	int result, size;
	unsigned long cmtline;
	unsigned eventmask = iot->eventmask;

	if( eventmask&eventflag ) {
		if( iot->event( iob, iot, startevent,
				(void __F*)cmt )==IOTEXT_DISPATCH_FAIL )
			return IOTEXT_DISPATCH_FAIL;
		eventmask = iot->eventmask;
	}

	result = IO_END;
	cmtline=0;

	do {

		_integer1 __F* p = iob->lock( iob->context, &size, 0, IO_FLAG_READ );

		if( p==NULL ) {
			result = FAIL_IO;
			break;
		}

		if( size<=0 ) {
			if( !cmt[0] && lcmt ) result = IOTEXT_COMMENT;
			else result = IOTEXT_INVALID_COMMENT;
			iob->unlock( iob->context, p );
			break;
		}

		while( size>0 ) {

			char c = *p;

			if( c=='\n' || c=='\r' ) {

				cmtline++;
				size--;

				p = iotext_skip_lineend( iob, p+1, &size );

				if( !p ) {
/* XXX error code for this case can be IOTEXT_INVALID_COMMENT - consider this ! */
					result = FAIL_IO;
					break;
				}

				if( !cmt[0] ) {
					result = IOTEXT_COMMENT;
					break;
				}

				/* XXX parsing for ' # endif' support */
				c = IOTEXT_LINE_END;

			}
			else {

				if( *cmt==c ) {

					int cmp = iotext_compare_string(iob,iot,&p,&size,cmt,lcmt);

					if( cmp==FAIL_UNAVAILABLE ) {
						/* insufficient buffer means trailing specifier was
						  not found */
						result = IOTEXT_INVALID_COMMENT;
						break;
					}

					if( cmp==0 ) {
						DEBUG_PUTS(( "iotext_skip_comment:a comment found" ));
						p+=lcmt;
						result = IOTEXT_COMMENT;
						break;
					}

				} /* end if( *cmt==c ) */

				size--;
				p++;

			} /* end if( line end is found ) */

			if( eventmask&charflag ) {
				if( iot->event(iob,iot,charevent,&c)==IOTEXT_DISPATCH_FAIL ) {
					result = IOTEXT_DISPATCH_FAIL;
					break;
				}
				eventmask = iot->eventmask; /* update event mask */
			}
		} /* end while( locked chars ) */

		iob->unlock( iob->context, p );

	} while( result==IO_END );

	if( result==IOTEXT_COMMENT ) {
		if( eventmask&eventflag ) {
			if( iot->event( iob, iot, endevent, size<=0?
					NULL:(void __F*)cmt )==IOTEXT_DISPATCH_FAIL )
				result = IOTEXT_DISPATCH_FAIL;
		}
	}

	iot->invalidcmt = *line;
	*line += cmtline;
	return result;
}

__MODULEEND__



__MODULEBEGIN__( "iotread.c" )
/* iotextRead */

static int
iotext_read
	P_B(( iob, token, tokenlen, buf, bufsize, flag, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		const char __F* token
	P_N		int tokenlen
	P_N		char __F* buf
	P_N		int bufsize
	P_N		unsigned flag
	P_N		struct _IOTEXT __F* iot
	P_E
{
	struct _IOBUFFER iobcache;
	_integer1 __F* p;
	unsigned long line;
	unsigned eventmask;

	char c, __F* curbuf;
	int i, result=IO_END, len=0;

	OP_COPY( iobcache, *iob );
	curbuf = buf;

	/*  following routine must be done after the above routine
	   because the current line can be modified by it */
	line = iot->line;
	eventmask = iot->eventmask;

	while( 1 ) {

		/* lock available buffer */
		int size;
		p = iobcache.lock( iobcache.context, &size, 0, IO_FLAG_READ );
		if( !p ) return FAIL_IO;

		/* no locked chars, terminate loop */
		if( size==0 ) break;

		/* parse locked contents */
		while( size>0 ) {

			c = *p;

			if( !(flag&IOTEXT_FLAG_DISABLE_COMMENT) ) {
				int lcmt;
				const char __F* cmt;

				/* line comment */
				lcmt = iot->llinecmt;
				cmt = iot->linecmt;
				if( lcmt && *cmt==c &&
				    iotext_compare_string(iob,iot,&p,&size,cmt,lcmt)==0 ) {
					iobcache.unlock( iobcache.context, p+lcmt );
					result = iotext_skip_comment( iob, iot,
								"", 1,
								IOTEXT_FLAG_LINE_COMMENT,
								IOTEXT_FLAG_LINE_COMMENT_CHAR,
								IOTEXT_BEGIN_LINE_COMMENT,
								IOTEXT_END_LINE_COMMENT,
								IOTEXT_LINE_COMMENT_CHAR,
								&line );
					goto l_result;
				}


				lcmt = iot->lblockcmt1;
				cmt = iot->blockcmt1;
				if( lcmt && *cmt==c &&
				    iotext_compare_string(iob,iot,&p,&size,cmt,lcmt)==0 ) {
					iobcache.unlock( iobcache.context, p+lcmt );
					result = iotext_skip_comment( iob, iot,
								iot->blockcmt2, iot->lblockcmt2,
								IOTEXT_FLAG_BLOCK_COMMENT,
								IOTEXT_FLAG_BLOCK_COMMENT_CHAR,
								IOTEXT_BEGIN_BLOCK_COMMENT,
								IOTEXT_END_BLOCK_COMMENT,
								IOTEXT_BLOCK_COMMENT_CHAR,
								&line );
					goto l_result;
				}

				lcmt = iot->lblockcmt2;
				cmt = iot->blockcmt2;
				if( lcmt && *cmt==c &&
				    iotext_compare_string(iob,iot,&p,&size,cmt,lcmt)==0 ) {
					iobcache.unlock(iobcache.context,p+lcmt);
					/* invalid block comment : trailing specifier is found
					   without leading specifier */
					iot->invalidcmt = line;
					result = IOTEXT_INVALID_COMMENT;
					goto l_result;
				}
			} /* end if( !(flag&IOTEXT_FLAG_DISABLECOMMENT) ) */

			size--;
			p++;

			if( !(flag&IOTEXT_FLAG_PRESERVE_LINE_END) &&
				(c=='\n' || c=='\r') ) {
				/*  io_skip_lineend should be called after finding
				   exclusive mode delimiter because buffer can be regressed
				   in that routine */
				line++;
				c = IOTEXT_LINE_END;
			}

			/* progress locked char, must be here not to be ignored by
				executing unlock */

			if( !( flag&IOTEXT_FLAG_DISABLE_DELIMITER ) &&
			     ( flag&IOTEXT_FLAG_EXCLUSIVE?
			             memchr(token,c,tokenlen)==NULL :
			             memchr(token,c,tokenlen)!=NULL ) ) {
				iot->lastchar = c;

				if( flag&IOTEXT_FLAG_EXCLUSIVE ) {
					/* ignore non-token character */
					size++;
					p--;
					if( c==IOTEXT_LINE_END ) line--;
				}
				else {
					/* post event
					    IOTEXT_DELIMITER is not posted for the exclusive token
					    because it is ignored and IOTEXT_CHAR will be
					    posted for that character in the future - one event
					    for one character */
					if( (eventmask&IOTEXT_FLAG_DELIMITER) &&
					    iot->event(iob,iot,IOTEXT_DELIMITER,&c)==
					    	IOTEXT_DISPATCH_FAIL )
						result = IOTEXT_DISPATCH_FAIL;

					if( c==IOTEXT_LINE_END ) {
						p = iotext_skip_lineend( iob, p, &size );
						if( !p ) {
							result = FAIL_IO;
							goto l_result;
						}
					}
				}

				/* escape duplicated loops */
				result = IOTEXT_DELIMITER;
				goto l_terminate;
			}

			if( c==IOTEXT_LINE_END ) {
				p = iotext_skip_lineend( iob, p, &size );
				if( !p ) {
					result = FAIL_IO;
					goto l_result;
				}
			}


			/* post event : a meaningful character was read */
			if( eventmask&IOTEXT_FLAG_CHAR ) {
				switch( iot->event( iob, iot, IOTEXT_CHAR, &c ) ) {
					/* currently the buffer is locked so unlocking is needed */
					case IOTEXT_DISPATCH_FAIL :
					result = IOTEXT_DISPATCH_FAIL;
					goto l_terminate;

					case IOTEXT_SKIP_CHAR : continue;
				}
				eventmask = iot->eventmask; /* update event mask */
			}

			/* store the character if needed */
			if( curbuf ) {
				if( bufsize>len ) {
					*curbuf++ = c;
				}
				else if( !(flag&IOTEXT_FLAG_FORCE) ) {
					/* buffer is too short and flag to force not to stop
					   until a token is found was not specified */
					result = FAIL_BUFFER;
					goto l_terminate;
				}
			}

			/* increase read length */
			len++;

		} /* end while( iteration for locked chars ) */

		iobcache.unlock( iobcache.context, p );
	} /* end while(1) */


	/* post process to report results */

	l_terminate:
	iobcache.unlock( iobcache.context, p );

	l_result:

	/* terminate string if its buffer was supplied */
	if( curbuf ) {
		*curbuf = 0;
		/* trailing blanks were not wanted */
		if( !(flag&IOTEXT_FLAG_KEEP_TRAIL_BLANK) ) {
			i = strBackwardTokenIndex( curbuf, len, str_white_char,
				STR_FLAG_END_OF_STRING|STR_FLAG_EXCLUSIVE );
			if( i>=0 ) {
				buf[i] = 0;
				len = i;
			}
		}
	} /* end if( curbuf ) */

	iot->line = line;
	iot->length = len;

	return result;
}



int __API
iotextRead
	P_B(( iob, token, buf, bufsize, flag, iot ))
P_1	struct _IOBUFFER __F* iob
P_N	const char __F* token
P_N	char __F* buf
P_N	int bufsize
P_N unsigned flag
P_N struct _IOTEXT __F* iot
	P_E
{
	int result, wclen, tokenlen;
	char wcbuf[OP_ELEM_COUNT(str_white_char)];
	const char __F * wc;

	bufsize--; /* reserve a byte for string termination */

	/* set default token if there is no given token */
	if( !token ) token = str_white_char;

	if( !iot ) { /* set default io text data if there is no given data */
		iot = &io_text_default;
		OP_ENABLE( flag, IOTEXT_FLAG_DISABLE_COMMENT );
	}

	/* initialize white character token for initial skipping */

	if( flag&IOTEXT_FLAG_KEEP_LEAD_BLANK ) {
		/* formal initialization not to cause compiler warning */
		wc = NULL;
		wclen = 0;
	}
	else {
		if( flag&IOTEXT_FLAG_KEEP_LEAD_DELIMITER ) {
			/* exclude white characters in the supplied tokens */
			int i;
			char c, __F* curbuf=wcbuf;
			for( i=0, wclen=0; i<OP_ELEM_COUNT(wcbuf); i++ ) {
				/*  the current white character is not found in the token,
				   so add it */
				if( !strchr(token,c=str_white_char[i]) ) {
					*curbuf++ = c;
					wclen++;
				}
			}
			*curbuf=0; /* XXX maybe not needed because memchr is used currently - inspect it */
			wc = wcbuf;
		}
		else {
			wc = str_white_char;
			wclen = OP_ELEM_COUNT(str_white_char)-1;
		}
	}

	tokenlen = strlen(token);
	while( 1 ) {
		if( !(flag&IOTEXT_FLAG_KEEP_LEAD_BLANK) ) {
			result = iotext_read( iob, wc, wclen, NULL, 0,
				IOTEXT_FLAG_EXCLUSIVE|IOTEXT_FLAG_DISABLE_COMMENT, iot );
			/* if return value is IO_END, the given string buffer should be
			  terminated, so result==0 is excluded */
			if( result<0 ) break;
		}

		result = iotext_read(iob,token,tokenlen,buf,bufsize,flag,iot);
		if( result!=IOTEXT_COMMENT || iot->length>0 ||
			(flag&IOTEXT_FLAG_KEEP_LEAD_COMMENT) ) break;
	}

	return result;
}

__MODULEEND__



#if 0
__MODULEBEGIN__( "iotrdstr.c" )
/* iotextReadString

  leadchar can be zero if it is not present - evidently possible case - but
 trailchar must not be zero */

int __API
iotextReadString
	P_B(( iob, bufptr, bufsize, lead, trail, lastchar, flag, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		char* bufptr
	P_N		int bufsize
	P_N		const char __F* lead
	P_N		const char __F* trail
	P_N		char lastchar
	P_N		unsigned flag
	P_N		struct _IOTEXT __F* iot
	P_E
{
	struct _IOBUFFER iobcache;
	int result, len;
	char token[5];
	unsigned iotrdflag;

	/* set default data if data was not supplied */
	if( !iot ) iot = &io_text_default;

#if 0
	/* if there is leading string specifier and
		leading string specifier was not found */
	if( leadchar  && leadchar!=lastchar ) {
		token[0] = leadchar; /* wait leading string specifier without garbage */
		token[1] = 0;
		result = iotextWait( iob, token, iot );
		if( result<=0 )	return result;
	}
#endif
	if( lead ) {
		iotext_compare_string( iob, iot, locked, remained, s, len );
	}

	/* skip leading blanks if requested */
	if( flag&IOTEXT_FLAG_SKIP_LEAD_BLANK )
		iotextObserve( iob, NULL, iot );

	/* find trailing string specifier */

	token[0] = '\n';
	token[1] = '\r';
	token[2] = '\\';
	token[3] = trailchar;
	token[4] = 0;

	len = 0;
	iotrdflag = flag&(IOTEXT_FLAG_PRESERVE_LINE_END|IOTEXT_FLAG_FORCE);
	OP_COPY( iobcache, *iob );

	if( !(flag&IOTEXT_FLAG_ENABLE_COMMENT) )
		OP_ENABLE( iotrdflag, IOTEXT_FLAG_DISABLE_COMMENT );

	if( !(flag&IOTEXT_FLAG_SKIP_LEAD_BLANK) )
		OP_ENABLE( iotrdflag, IOTEXT_FLAG_KEEP_LEAD_BLANK );

	if( !(flag&IOTEXT_FLAG_SKIP_TRAIL_BLANK) )
		OP_ENABLE( iotrdflag, IOTEXT_FLAG_KEEP_TRAIL_BLANK );

	while( (result=iotextRead( iob, token, bufptr, bufsize,
					iotrdflag, iot ))>0 ) {

		if( result!=IOTEXT_COMMENT ) {
			/* evidently returned value is IOTEXT_DELIMITER */
			register int l = iot->length;
			len += l;
			if( bufsize>0 ) {
				bufptr += l;
				bufsize -= l;
			}

//???? unneeded codes
//		if( result==IOTEXT_COMMENT ) {
//			if( trailchar=='\n' ) break;
//result = result;
//		}
///*		else*/ {
			lastchar = (char)iot->lastchar;

			if( lastchar==trailchar ) {
				/* skipping trailing blanks before the trailing character
					was requested */
				if( flag&IOTEXT_FLAG_SKIP_TRAIL_BLANK ) {
					strBackwardToken( bufptr, len, str_white_char,
						STR_FLAG_END_OF_STRING|STR_FLAG_EXCLUSIVE );
				} /* end if( flag&IOTEXT_FLAG_SKIP_TRAIL_BLANK ) */

				iot->length = len;
				return 1; /* success */
			}
			else { /* lastchar!=trailchar */
				char __F* p;

				switch( lastchar ) {

					case '\\' :
					p = (char __F*) iobcache.lock
							( iobcache.context, NULL, 2, IO_FLAG_READ );
					if( p ) {
						char c=*p++;
						/* \ is the line continuation character */
						if( c=='\n' || c=='\r' ) {
							if( flag&IOTEXT_FLAG_NO_LINE_CONT ) {
								goto l_plain_bkslash;
							}
							else {
								if( c=='\n' ) {
									/* if cr followed, eat it */
									if( *p=='\r' ) {
										p++;
										lastchar = '\r';
									}
									else {
										lastchar = '\n';
									}
								}
								else {
									/* if lf followed, eat it */
									if( *p=='\n' ) {
										p++;
										lastchar = '\n';
									}
									else {
										lastchar = '\r';
									}
								} /* end if( c=='\n' ) */

							} /* end if( flag&IOTEXT_FLAG_NO_LINE_CONT ) */
						}
						else {
							/* \ is the character escape sequence */

							if( flag&IOTEXT_FLAG_RESERVE_CHAR_ESC ) {
								l_plain_bkslash:
								len+=2;
								if( bufsize>2 ) {
									*bufptr++ = '\\';
									*bufptr++ = c;
									bufsize-=2;
								}
								else if( !(flag&IOTEXT_FLAG_FORCE) )
									return FAIL_BUFFER;
							}
							else {
								switch( c ) {
									case 'n'	: c = '\n'; break;
									case 'r'	: c = '\r'; break;
									case 't'	: c = '\t'; break;
									case '\\' : c = '\\'; break;
									case '0'	: c = 0;
								}

								iot->length++;
								if( bufsize>1 ) {
									*bufptr++=c;
									bufsize--;
								}
								else if( !(flag&IOTEXT_FLAG_FORCE) ) {
									return FAIL_BUFFER;
								}
							}
						} /* end if( c=='\n' || c=='\r' ) */
						iobcache.unlock( iobcache.context, p );
					}
					else { /* iob_lock is failed */
						return FAIL_IO;
					}
					break;

					case '\r' :
					case '\n' : /* a hint, the last character cannot be '\r' if iotextRead was used  XXX */
					if( flag&IOTEXT_FLAG_ALLOW_LINE_END ) {
						len++;
						if( bufsize>1 ) {
							*bufptr++ = lastchar;
							bufsize--;
						}
						else if( !(flag&IOTEXT_FLAG_FORCE) ) {
							return FAIL_BUFFER;
						}
					}
					else {
						return IOTEXT_INVALID_STRING;
					}

				} /* end switch( lastchar ) */

			} /* end if( lastchar==trailchar ) */

		} /* end if( result==IOTEXT_COMMENT ) */

	} /* end while( iotextRead ) */

	/* failure */
	if( leadchar && result==IO_END ) return IOTEXT_INVALID_STRING;
	return result;
}
#endif
__MODULEEND__
#if 1
int __API
iotextReadString
	P_B(( iob, bufptr, bufsize, leadchar, trailchar, lastchar, flag, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		char* bufptr
	P_N		int bufsize
	P_N		char leadchar
	P_N		char trailchar
	P_N		char lastchar
	P_N		unsigned flag
	P_N		struct _IOTEXT __F* iot
	P_E
{
	struct _IOBUFFER iobcache;
	int result, len;
	char token[5];
	unsigned iotrdflag;

	/* set default data if data was not supplied */
	if( !iot ) iot = &io_text_default;

	/* if there is leading string specifier and
		leading string specifier was not found */
	if( leadchar  && leadchar!=lastchar ) {
		token[0] = leadchar; /* wait leading string specifier without garbage */
		token[1] = 0;
		result = iotextWait( iob, token, iot );
		if( result<=0 )	return result;
	}

	/* skip leading blanks if requested */
	if( flag&IOTEXT_FLAG_SKIP_LEAD_BLANK )
		iotextObserve( iob, NULL, iot );

	/* find trailing string specifier */

	token[0] = '\n';
	token[1] = '\r';
	token[2] = '\\';
	token[3] = trailchar;
	token[4] = 0;

	len = 0;
	iotrdflag = flag&(IOTEXT_FLAG_PRESERVE_LINE_END|IOTEXT_FLAG_FORCE);
	OP_COPY( iobcache, *iob );

	if( !(flag&IOTEXT_FLAG_ENABLE_COMMENT) )
		OP_ENABLE( iotrdflag, IOTEXT_FLAG_DISABLE_COMMENT );

	if( !(flag&IOTEXT_FLAG_SKIP_LEAD_BLANK) )
		OP_ENABLE( iotrdflag, IOTEXT_FLAG_KEEP_LEAD_BLANK );

	if( !(flag&IOTEXT_FLAG_SKIP_TRAIL_BLANK) )
		OP_ENABLE( iotrdflag, IOTEXT_FLAG_KEEP_TRAIL_BLANK );

	while( (result=iotextRead( iob, token, bufptr, bufsize,
					iotrdflag, iot ))>0 ) {

		if( result!=IOTEXT_COMMENT ) {
			/* evidently returned value is IOTEXT_DELIMITER */
			register int l = iot->length;
			len += l;
			if( bufsize>0 ) {
				bufptr += l;
				bufsize -= l;
			}

//???? unneeded codes
//		if( result==IOTEXT_COMMENT ) {
//			if( trailchar=='\n' ) break;
//result = result;
//		}
///*		else*/ {
			lastchar = (char)iot->lastchar;

			if( lastchar==trailchar ) {
				/* skipping trailing blanks before the trailing character
					was requested */
				if( flag&IOTEXT_FLAG_SKIP_TRAIL_BLANK ) {
					strBackwardToken( bufptr, len, str_white_char,
						STR_FLAG_END_OF_STRING|STR_FLAG_EXCLUSIVE );
				} /* end if( flag&IOTEXT_FLAG_SKIP_TRAIL_BLANK ) */

				iot->length = len;
				return 1; /* success */
			}
			else { /* lastchar!=trailchar */
				char __F* p;

				switch( lastchar ) {

					case '\\' :
					p = (char __F*) iobcache.lock
							( iobcache.context, NULL, 2, IO_FLAG_READ );
					if( p ) {
						char c=*p++;
						/* \ is the line continuation character */
						if( c=='\n' || c=='\r' ) {
							if( flag&IOTEXT_FLAG_NO_LINE_CONT ) {
								goto l_plain_bkslash;
							}
							else {
								if( c=='\n' ) {
									/* if cr followed, eat it */
									if( *p=='\r' ) {
										p++;
										lastchar = '\r';
									}
									else {
										lastchar = '\n';
									}
								}
								else {
									/* if lf followed, eat it */
									if( *p=='\n' ) {
										p++;
										lastchar = '\n';
									}
									else {
										lastchar = '\r';
									}
								} /* end if( c=='\n' ) */

							} /* end if( flag&IOTEXT_FLAG_NO_LINE_CONT ) */
						}
						else {
							/* \ is the character escape sequence */

							if( flag&IOTEXT_FLAG_RESERVE_CHAR_ESC ) {
								l_plain_bkslash:
								len+=2;
								if( bufsize>2 ) {
									*bufptr++ = '\\';
									*bufptr++ = c;
									bufsize-=2;
								}
								else if( !(flag&IOTEXT_FLAG_FORCE) )
									return FAIL_BUFFER;
							}
							else {
								switch( c ) {
									case 'n'	: c = '\n'; break;
									case 'r'	: c = '\r'; break;
									case 't'	: c = '\t'; break;
									case '\\' : c = '\\'; break;
									case '0'	: c = 0;
								}

								iot->length++;
								if( bufsize>1 ) {
									*bufptr++=c;
									bufsize--;
								}
								else if( !(flag&IOTEXT_FLAG_FORCE) ) {
									return FAIL_BUFFER;
								}
							}
						} /* end if( c=='\n' || c=='\r' ) */
						iobcache.unlock( iobcache.context, p );
					}
					else { /* iob_lock is failed */
						return FAIL_IO;
					}
					break;

					case '\r' :
					case '\n' : /* a hint, the last character cannot be '\r'
									 if iotextRead was used  XXX */
					if( flag&IOTEXT_FLAG_ALLOW_LINE_END ) {
						len++;
						if( bufsize>1 ) {
							*bufptr++ = lastchar;
							bufsize--;
						}
						else if( !(flag&IOTEXT_FLAG_FORCE) ) {
							return FAIL_BUFFER;
						}
					}
					else {
						return IOTEXT_INVALID_STRING;
					}

				} /* end switch( lastchar ) */

			} /* end if( lastchar==trailchar ) */

		} /* end if( result==IOTEXT_COMMENT ) */

	} /* end while( iotextRead ) */

	/* failure */
	if( leadchar && result==IO_END ) return IOTEXT_INVALID_STRING;
	return result;
}

__MODULEEND__
#endif


__MODULEBEGIN__( "iotsleep.c" )
/* iotextSleep

 Sleep until a token is found : garbage is allowed

  IOTEXT_FLAG_KEEP_LEAD_BLANK is applied for iotextSleep on
 2000.05.31 - actually bug fix for the case, for example, "\n" is specified
 as a token but the initial skipping routine in iotextRead will kill a '\n'
 which must not be wasted at there and another line which must not be
 skipped will be skipped. */

int __API iotextSleep
	P_B(( iob, token, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		const char __F* token
	P_N		struct _IOTEXT __F* iot
	P_E
{
	int result;

	while( (result=iotextRead( iob, token, NULL, 0,
					 IOTEXT_FLAG_KEEP_LEAD_BLANK, iot ))==IOTEXT_COMMENT ) {
	}

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "iotwait.c" )
/* iotextWait

 Wait until a token is found : garbage is not allowed */

int __API iotextWait
	P_B(( iob, token, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		const char __F* token
	P_N		struct _IOTEXT __F* iot
	P_E
{
	int result;
	struct _IOTEXT iotdef;

	/* set default data if there is no given data */
	if( !iot ) {
		mem_short_set( &iotdef, 0, sizeof(iotdef) );
		iot = &iotdef;
	}

	/* ignore the case which returned by comment */
	while( (result=iotextRead( iob, token, NULL, 0,
					 IOTEXT_FLAG_KEEP_LEAD_DELIMITER, iot ))==IOTEXT_COMMENT ) {
		if( iot->length>0 )
			return IOTEXT_GARBAGE_FOUND;
	}

	if( result>=0 && iot->length>0 )
		return IOTEXT_GARBAGE_FOUND;

	return result;
}
__MODULEEND__



__MODULEBEGIN__( "iotskipl.c" )
/* iotextSkipLine */

int __API
iotextSkipLine
	P_B(( iob, n, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		unsigned long n
	P_N		struct _IOTEXT __F* iot
	P_E
{
	int result=FAIL_IO;
	while( n-->0 ) {
		result = iotextSleep(iob,str_lineend_char,iot);
		if( result<=0 ) break;
	}
	return result;
}

__MODULEEND__



__MODULEBEGIN__( "iotobsrv.c" )
/* iotextObserve

 Observe specified tokens. */

int __API iotextObserve
	P_B(( iob, token, iot ))
	P_1		struct _IOBUFFER __F* iob
	P_N		const char __F* token
	P_N		struct _IOTEXT __F* iot
	P_E
{
	int result;
	while( (result=iotextRead( iob,token, NULL, 0,
	                 IOTEXT_FLAG_KEEP_LEAD_DELIMITER|
	                 IOTEXT_FLAG_EXCLUSIVE, iot ))==IOTEXT_COMMENT );
	return result;
}
__MODULEEND__



__MODULEBEGIN__( "iotconv.c" )
/* iotextConvert

 Returns source format.

 originated from tounix utility
 1998. 10. 9  DOS text to UNIX text conversion
 1998. 10. 25 UNIX text to DOS text conversion
 Hee S. Chung */

unsigned __API
iotextConvert
	P_B(( dest, src, endline ))
	P_1		struct _IO __F* dest
	P_N		struct _IO __F* src
	P_N		unsigned endline
	P_E
{
	_integer1 c;
	long n=0;
	unsigned result=0, leader=0;
	char lineend=0;

	char __F* str= (char __F*)src;

	/* string support on 2000.03.13 using internal flag for iotextWrite */

	while( endline&_IOTEXT_FLAG_MEMORY?
				 (c=*str++)!=0 : (n=ioRead(src,&c,1,0))==1 ) {

		switch( c ) {
			_integer1 cr, lf;

			/* if a line end was found */
			case '\r' :
			result |= IOTEXT_CR;
			if( !leader ) leader=IOTEXT_CR;
			goto lineend_found;

			case '\n' :
			result |= IOTEXT_LF;
			if( !leader ) leader=IOTEXT_LF;

			lineend_found:
			/* line end char was not found or found line end char is
				appeared again, that means successive new line is found
				without non-blank chars, bug fixed 1999.05.28 */
			if( !lineend || lineend==c ) {
				/* consider inverted case */
				if( endline&0x4 )
					cr='\n', lf='\r';
				else
					cr='\r', lf='\n';

				/* write line end chars */
				if( endline&IOTEXT_CR )
					if( !ioFullWrite(dest,&cr,1,0) ) return 0;
				if( endline&IOTEXT_LF )
					if( !ioFullWrite(dest,&lf,1,0) ) return 0;

				/* currently accepted line end */
				lineend = c;
			}
			break;

			default :
			/* line end char was not found */
			lineend = 0;
			/* just write the read byte to destination */
			if( !ioFullWrite(dest,&c,1,0) ) return 0;
		}
	}

	if( n==FAIL_IO ) return 0;

	return
		/* if result was not modified */
		result==0? IOTEXT_UNKNOWN :
		/* if lf cr found */
		result==IOTEXT_CRLF && leader==IOTEXT_LF? IOTEXT_LFCR :
		/* plain case */
		result;
}

__MODULEEND__



__MODULEBEGIN__( "iotfrmat.c" )
/* iotextFormat */

unsigned __API iotextFormat
	P_B(( src, readbytes ))
	P_1		struct _IO __F* src
	P_N		long __F* readbytes
	P_E
{
	char c, next;
	long n, total=0;

	while( (n=ioRead(src,&c,1,0))==1 ) {

		total++;

		/* check line end */
		switch( c ) {
			case '\r' :
			case '\n' :
			n = ioRead(src,&next,1,0);
			if( n==FAIL_IO ) return 0;
			if( readbytes ) *readbytes = total + n;
			if( c=='\r' ) {
				if( n==0 || next!='\n' ) return IOTEXT_CR;
				return IOTEXT_CRLF;
			}
			else {
				if( n==0 || next!='\r' ) return IOTEXT_LF;
				return IOTEXT_LFCR;
			}
		} /* end switch( currently read char ) */
	} /* end while( a byte can be read from src ) */

	if( readbytes )
		*readbytes = total;

	return n==FAIL_IO ? 0 : IOTEXT_UNKNOWN;
}

__MODULEEND__



__MODULEBEGIN__( "iotwrite.c" )
/* iotextWrite

 Returns string format or FAIL_IO if io operation was failed. */

_bool __API
iotextWrite
	P_B(( io, buf, endline ))
	P_1		struct _IO __F* io
	P_N		const char __F* buf
	P_N		unsigned endline
	P_E
{
	/* call iotextConvert with _IOTEXT_FLAG_MEMORY */
	return iotextConvert( io, (struct _IO __F*)buf, endline|_IOTEXT_FLAG_MEMORY )>0;
}
__MODULEEND__



__MODULEBEGIN__( "iotwstr.c" )
/* iotextWriteString */

_bool __API
iotextWriteString
	P_B(( io, s ))
	P_1		struct _IO __F* io
	P_N		const char __F* s
	P_E
{
	/* no byte ordering will be occurred so casting to char* is valid */
	return ioFullWrite( io, (void __F*)s, strlen(s), 0 );
}

__MODULEEND__


/* end of iotext.c */

